/*
 * Copyright 2006 Google Inc.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.google.gwt.user.client.ui;

import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.Element;
import com.google.gwt.user.client.Event;

/**
 * A widget that presents a list of choices to the user, either as a list box or
 * as a drop-down list.
 * 
 * <p>
 * <img class='gallery' src='ListBox.png'/>
 * </p>
 * 
 * <h3>CSS Style Rules</h3>
 * <ul class='css'>
 * <li>.gwt-ListBox { }</li>
 * </ul>
 * 
 * <h3>Example</h3>
 * {@link com.google.gwt.examples.ListBoxExample code}
 */
public class ListBox extends FocusWidget implements SourcesChangeEvents {

  private ChangeListenerCollection changeListeners;

  /**
   * Creates an empty list box.
   */
  public ListBox() {
    super(DOM.createSelect());
    sinkEvents(Event.ONCHANGE);
  }

  public void addChangeListener(ChangeListener listener) {
    if (changeListeners == null)
      changeListeners = new ChangeListenerCollection();
    changeListeners.add(listener);
  }

  /**
   * Adds an item to the list box.
   * 
   * @param item the text of the item to be added
   */
  public void addItem(String item) {
    insertItem(item, getItemCount());
  }

  /**
   * Removes all items from the list box.
   */
  public void clear() {
    Element h = getElement();
    while (DOM.getChildCount(h) > 0)
      DOM.removeChild(h, DOM.getChild(h, 0));
  }

  /**
   * Gets the number of items present in the list box.
   * 
   * @return the number of items
   */
  public int getItemCount() {
    return DOM.getChildCount(getElement());
  }

  /**
   * Gets the text associated with the item at the specified index.
   * 
   * @param index the index of the item whose text is to be retrieved
   * @return the text associated with the item
   */
  public String getItemText(int index) {
    Element child = DOM.getChild(getElement(), index);
    return DOM.getInnerText(child);
  }

  /**
   * Gets the currently-selected item. If multiple items are selected, this
   * method will return the first selected item ({@link #isItemSelected(int)}
   * can be used to query individual items).
   * 
   * @return the selected index, or <code>-1</code> if none is selected
   */
  public int getSelectedIndex() {
    return DOM.getIntAttribute(getElement(), "selectedIndex");
  }

  /**
   * Determines whether an individual list item is selected.
   * 
   * @param itemIndex the index of the item to be tested
   * @return <code>true</code> if the item is selected
   */
  public boolean isItemSelected(int itemIndex) {
    Element select = DOM.getChild(getElement(), itemIndex);
    if (select == null)
      throw new IndexOutOfBoundsException();
    return DOM.getAttribute(select, "selected").equals("true");
  }

  /**
   * Gets the number of items that are visible. If only one item is visible,
   * then the box will be displayed as a drop-down list.
   * 
   * @return the visible item count
   */
  public int getVisibleItemCount() {
    return DOM.getIntAttribute(getElement(), "size");
  }

  /**
   * Inserts an item into the list box.
   * 
   * @param item the text of the item to be inserted
   * @param idx the index at which to insert it
   */
  public void insertItem(String item, int idx) {
    Element option = DOM.createElement("OPTION");
    DOM.setInnerText(option, item);
    DOM.insertChild(getElement(), option, idx);
  }

  /**
   * Gets whether this list allows multiple selection.
   * 
   * @return <code>true</code> if multiple selection is allowed
   */
  public boolean isMultipleSelect() {
    return DOM.getAttribute(getElement(), "multiple").equals("true");
  }

  public void onBrowserEvent(Event event) {
    if (DOM.eventGetType(event) == Event.ONCHANGE) {
      if (changeListeners != null)
        changeListeners.fireChange(this);
    }
  }

  public void removeChangeListener(ChangeListener listener) {
    if (changeListeners != null)
      changeListeners.remove(listener);
  }

  /**
   * Removes the item at the specified index.
   * 
   * @param idx the index of the item to be removed
   */
  public void removeItem(int idx) {
    Element child = DOM.getChild(getElement(), idx);
    DOM.removeChild(getElement(), child);
  }

  /**
   * Sets whether this list allows multiple selections
   * 
   * @param multiple <code>true</code> to allow multiple selections
   */
  public void setMultipleSelect(boolean multiple) {
    DOM.setAttribute(getElement(), "multiple", multiple ? "true" : "false");
  }

  /**
   * Sets the currently selected index.
   * 
   * @param index the index of the item to be selected
   */
  public void setSelectedIndex(int index) {
    DOM.setIntAttribute(getElement(), "selectedIndex", index);
  }

  /**
   * Sets the number of items that are visible. If only one item is visible,
   * then the box will be displayed as a drop-down list.
   * 
   * @param visibleItems the visible item count
   */
  public void setVisibleItemCount(int visibleItems) {
    DOM.setIntAttribute(getElement(), "size", visibleItems);
  }
}
